#!/usr/bin/perl

# pmm2tb.pl

# Copyright (c) 2004, 2005 Robert Dahlem (pmm2tb@dahlem.name)

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


use POSIX;

$PMMAIL_SEP = "";

%dayname = (
	"0" => "Sun",
	"1" => "Mon",
	"2" => "Tue",
	"3" => "Wed",
	"4" => "Thu",
	"5" => "Fri",
	"6" => "Sat"
);

%monthname = (
	0 => "Jan",
	1 => "Feb",
	2 => "Mar",
	3 => "Apr",
	4 => "May",
	5 => "Jun",
	6 => "Jul",
	7 => "Aug",
	8 => "Sep",
	9 => "Oct",
	10 => "Nov",
	11 => "Dec"
);
	

# These are from:
# http://lxr.mozilla.org/mozilla/source/mailnews/base/public/nsMsgMessageFlags.h

$MOZ_FLAG_READ     = 0x0001; # has been read
$MOZ_FLAG_REPLIED  = 0x0002; # a reply has been successfully sent
$MOZ_FLAG_MARKED   = 0x0004; # the user-provided mark
$MOZ_FLAG_EXPUNGED = 0x0008; # deleted
$MOZ_FLAG_HAS_RE   = 0x0010; # whether subject has "Re:" on the front.
$MOZ_FLAG_ATT      = 0x10000000; # message has files attached to it

# Format FOLDER.BAG
# Field  0 : 0=unread, 1=read, 2=reply, 3=unsend, 4=draft
# Field  1 : 0=no Attachments, 1=with Attachments
# Field  2 : Date
# Field  3 : Time
# Field  4 : Subject
# Field  5 : To Email 
# Field  6 : To Name
# Field  7 : From Email
# Field  8 : From Name
# Field  9 : Size
# Field 10 : Filename
# Field 11 : Description of Attachments
# Field 12 : ??? number

$PMM_FLAG_UNREAD   = 0;
$PMM_FLAG_READ     = 1;
$PMM_FLAG_REPLIED  = 2;
$PMM_FLAG_UNSENT   = 3;
$PMM_FLAG_DRAFT    = 4;
$PMM_PRIO_HIGH     = 10;
$PMM_PRIO_LOW      = 100;

sub no_slashes {
	my ($s) = @_;

	$s =~ s!/!-!g;
	
	return $s;
}

# make Thunderbird folder
sub mk_folder {
	my ($folder_name) = @_;
	
	my $DUMMY;
	
	# make empty mbox
	open(DUMMY, ">", "$folder_name") or die "create $folder_name: $!";
	close(DUMMY);
	
	# make empty Thunderbird summary file
	open(DUMMY, ">", "$folder_name.msf") or die "create $folder_name.msf: $!";
	close(DUMMY);
	
	# make directory for subfolders
	-d "$folder_name.sbd" or mkdir "$folder_name.sbd" or die "mkdir $folder_name.sbd: $!";
}

# copy all mails from PMMail folder to Thunderbird folder
sub mig_folder {
	my ($pmmail_dir, $mbox) = @_;
	
	my $FOLDER_BAG;
	my $MSG;
	my $MBOX;
	my @bag_fields;
	my $filename;
	my $line;
	my %timestamp;
	my %flags;
	my %attached;
	
	mk_folder "$mbox";
	
	# read folders directory of messages
	open(FOLDER_BAG, "$pmmail_dir/folder.bag") or return;
	while(<FOLDER_BAG>) {
		@bag_fields = split($PMMAIL_SEP, $_);
		$filename=$bag_fields[10];
		$timestamp{$filename} = "$bag_fields[2]:$bag_fields[3]";
		$flags{$filename}=($bag_fields[0] % $PMM_PRIO_HIGH) % $PMM_PRIO_LOW;
		$attached{$filename}=$bag_fields[1];
	}
	close(FOLDER_BAG);

	open(MBOX, ">", $mbox) or die "create $mbox: $!";
	# sort messages by timestamp, dump to mbox format
	for my $i ( sort { $timestamp{$a} cmp $timestamp{$b} } keys %timestamp ) {
		if(!open(MSG, "$pmmail_dir/$i") ) {
			printf STDERR "read $pmmail_dir/$i: $!\n";
		} else {
			# Build From_ line
			($year, $mon, $mday, $hour, $min, $sec) = split(/[-:]/, $timestamp{$i});
			$time=POSIX::mktime($sec, $min, $hour, $mday, $mon-1, $year-1900);
			($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime($time);
			$from_str=sprintf("From - %s %s %02d %02d:%02d:%02d %04d",
				$dayname{$wday},
				$monthname{$mon},
				$mday,
				$hour, $minute, $sec, $year+1900);
			printf MBOX "%s\n", $from_str;
	
			# Build flags
			$mozilla_status=$MOZ_FLAG_READ;
			$mozilla_status2=0;
			if($flags{$i} == $PMM_FLAG_UNREAD) {
				$mozilla_status &= ~$MOZ_FLAG_READ;
			}
			if($flags{$i} == $PMM_FLAG_REPLIED) {
				$mozilla_status |= $MOZ_FLAG_REPLIED;
			}
			if($attached{$i}) {
				$mozilla_status2 |= $MOZ_FLAG_ATT;
			}
			printf MBOX "X-Mozilla-Status: %04.4x\n", $mozilla_status;
			printf MBOX "X-Mozilla-Status2: %08.8x\n", $mozilla_status2;
			
			# we built our own From_ line,
			# so suppress first line if it is From_ too
			$line = <MSG>;
			if(substr($line, 0, 5) ne "From ") {
				printf MBOX "%s", $line;
			}
				
			# copy rest of message
			while(<MSG>) {
				printf MBOX "%s%s",
					substr($_, 0, 5) eq "From "
						? " "
						: "",
					$_;
			}
			
			# make sure last line is empty
			printf MBOX "\n"  unless $_ eq "\n";
			 
			close(MSG);
		}
	}
	close(MBOX);
}

# recursively migrate all PMMail folders in $pmmail_dir
sub read_folder {
	my (
		$is_account_dir,	# Account Folder, no name
		$pmmail_dir,		# Directory to read from
		$pmmail_folder,		# PMMails folder name
		$mozilla_dir		# Directory to write to
	) = @_;
	
	my $pmmail_subfolder;
	my $mbox_name;
	my $FOLDER;
	my $FOLDER_INI;
	my %subfolder;

	if($is_account_dir) {
		$pmmail_subfolder = "";
	}else {
		# get PMMail folder name for this directory
		open(FOLDER_INI, "$pmmail_dir/folder.ini")
			or die "read $pmmail_dir/folder.ini: $!";
		($pmmail_subfolder) = split($PMMAIL_SEP, <FOLDER_INI>);
		close(FOLDER_INI);

		# make Thunderbird mbox name from PMMail folder name
		$mbox_name = no_slashes($pmmail_subfolder);
	
		printf "%s%s\n", $pmmail_folder, $mbox_name;

		# migrate all messages from this PMMail folder to Thunderbird mbox
		mig_folder($pmmail_dir, "$mozilla_dir/$mbox_name");
	}
	
	# read directory, gather PMMail subfolder names
	opendir(FOLDER, $pmmail_dir) or die "dir $pmmail_dir: $!";
	for my $folder ( grep { /\.fld$/i } readdir(FOLDER) ) {
		if(open(FOLDER_INI, "$pmmail_dir/$folder/folder.ini")) {
			($subfolder{$folder}) = split($PMMAIL_SEP, <FOLDER_INI>);
			close(FOLDER_INI);
		}
	}
	closedir(FOLDER);
	
	# sort by PMMail subfolder names, recurse into subfolders
	for my $pmmail_subdir ( sort { $subfolder{$a} cmp $subfolder{$b} } keys %subfolder ) {
		read_folder(
			0,
			"$pmmail_dir/$pmmail_subdir",
			"$pmmail_folder$pmmail_subfolder\\",
			$is_account_dir
				? "$mozilla_dir"
				: "$mozilla_dir/$mbox_name.sbd"
		);
	}
}

sub get_account_name {
	my ($account_dir) = @_;
	
	my $ACCT;
	
	open(ACCT, "$account_dir/acct.ini") or return;
	while(read(ACCT,$line,0x180) > 0) {
		($variable, $value) = split(/\0+/, substr($line, 1));
		if($variable eq "ACCTNAME") {
			close(ACCT);
			return $value;
		}
	}
	close(ACCT);
	return;	
}

sub read_accounts {
	my (
		$pmmail_dir,
		$mozilla_dir
	) = @_;
	
	my $ACCOUNTS;
	my $account_name;
	
	# read directory, gather PMMail account names
	opendir(ACCOUNTS, $pmmail_dir) or die "dir $pmmail_dir: $!";

	mk_folder "$mozilla_dir/PMMail";

	for my $account_dir ( grep { /\.act$/i } readdir(ACCOUNTS) ) {
		if(($account_name=get_account_name("$pmmail_dir/$account_dir"))) {
			$account_name=no_slashes($account_name);
			mk_folder("$mozilla_dir/PMMail.sbd/$account_name");
			read_folder(
				1,
				"$pmmail_dir/$account_dir",
				"PMMail\\$account_name",
				"$mozilla_dir/PMMail.sbd/$account_name.sbd");
		}
	}
	closedir(ACCOUNTS);
}	

# --- main ---
if($#ARGV ne 1) {
	printf STDERR "usage: %s PMMail_Account_Directory Thunderbird_Mail_Directory\n", $0;
	exit 1;
}
read_accounts($ARGV[0] eq "" ? "." : $ARGV[0], $ARGV[1]);
